from timemachines.skaters.tsa.tsaconstant import TSA_CONSTANT_SKATERS
from timemachines.skaters.tsa.tsaensembles import TSA_DO_ENSEMBLE_SKATERS
from timemachines.skaters.tsa.tsahypocratic import TSA_DO_HYPOCRATIC_SKATERS

TSA_SKATERS = TSA_CONSTANT_SKATERS + TSA_DO_ENSEMBLE_SKATERS + TSA_DO_HYPOCRATIC_SKATERS


from timemachines.skaters.flux.fluxinclusion import using_pyflux,pf
from timemachines.skatertools.utilities.conventions import Y_TYPE,A_TYPE, dimension
if using_pyflux:
    import pandas as pd
    from timemachines.skatertools.visualization.priorplot import prior_plot_exogenous
    from timemachines.skatertools.utilities.conventions import to_int_log_space, initialize_buffers, split_exogenous, update_buffers, \
        R_TYPE

    # TODO: Span models at https://pyflux.readthedocs.io/en/latest/gpnar.html with r

    def initialize_buffers(s, y: Y_TYPE, a: A_TYPE = None):  # FIXME: kill this
        s['n_obs'] = 0
        s['dim'] = dimension(y)
        s['a_dim'] = dimension(a)
        s['buffer'] = list()  # Target
        if s['dim'] > 1:
            s['exogenous'] = list()  # Exogenous, the rest of y
        s['model'] = None
        s['advance'] = list()  # Variables known in advance
        s['staleness'] = 0
        return s


    def update_buffers(s, a: A_TYPE, exog: [float], y0: float):
        # Store "target" and other observations or vars known in advance
        s['buffer'].append(y0)
        if exog is not None:
            s['exogenous'].append(exog)
        if a is not None:
            s['advance'].append(a)
        return s


    def flux_hyperparams(s, r: R_TYPE):
        # Interpret hyper-parameters - r is only used once at initiation of model
        integ_bounds, ar_bounds, ma_bounds =(0,2), (0.55, 2.45), (0.66, 10.33)
        s['n_burn'] = 25
        s['alpha'] = 0.25  # Defines confidence interval
        s['buffer_len'] = 5000
        s['integ'], s['ar'], s['ma'] = to_int_log_space(r, bounds=[integ_bounds, ar_bounds, ma_bounds])
        s['family'] = pf.Normal()
        return s


    def flux_or_last_value(s, k:int, exog:[float], y0:float):
        # We fit each time
        if s.get('model') is None:
            return y0, s, None
        else:
            y_hat = s['model'].predict(h=k)
            x = y_hat.values[0][0]
            w = None
            return x, s, w


    def flux_auto(y, s, k, a, t, e, r):
        """ One way to use flux package

                - Contemporaneous y[1:] variables are used as exogenous 'X' in pmdarima
                - This only works for k=1

            :returns: x, s', w
        """
        if s is None:
            s = dict()
            s = flux_hyperparams(s=s,r=r)
            s = initialize_buffers(s=s,y=y)

        if y is not None:
            # Process observation and return prediction
            assert isinstance(y, float) or len(y) == s['dim'], ' Cannot change dimension of input in flight '
            y0, exog = split_exogenous(y=y, dim=s['dim'])
            s = update_buffers(s=s, a=a, exog=exog, y0=y0)
            if True:  # Always fit prior to prediction
                none_, s, _ = flux_auto(y=None, s=s, k=k, a=a, t=t, e=e, r=r)  # Fit the model
                assert none_ is None
            return flux_or_last_value(s=s,k=k,exog=exog,y0=y0)

        if y is None:
            if len(s.get('buffer'))<s['n_burn']:
                s['model'] = None
            else:
                data = pd.DataFrame(columns=['y'], data=s.get('buffer'))
                s['model'] = pf.ARIMA(data=data, ar=s['ar'], ma=s['ma'], target='y', family=s['family'])
                _ = s['model'].fit("MLE")
            return None, s, None  # Acknowledge that a fit was requested by returning x=None, w=None


if __name__ == '__main__':
    assert using_pyflux,'pip install pyflux'
    err = prior_plot_exogenous(f=flux_auto, k=1, n=200, r=0.05)
    pass



from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any
import numpy as np
import math
from timemachines.skatertools.components import observance
from timemachines.skatertools.components.parade import parade


def regress_change_on_first_known(y:Y_TYPE, s:dict, k, a:A_TYPE=None, t:T_TYPE =None, e:E_TYPE =None )->([float] , Any , Any):
    """ Very basic modification of the last value cache.
        This looks at the contemporaneous influence of a single known in advance variable.
        Assumes independent increments when estimating the standard deviation.
        This is also intended to illustrate combination of skaters
    """
    y0 = wrap(y)[0]  # Ignore contemporaneous, exogenous variables
    if not s.get('prev_y0'):
        s = {'prev_y0':y0,
             'd':{}   # state for difference predicting skater
             }
        return y, 1.0, s
    else:
        dy0 = y0 - s['prev_y0']
        dy_hat, dy_hat_std = regress_level_on_first_known(y=[dy0], s=s['d'], k=k, a=a, t=t, e=e)
        x = [y0 + sum_dy for sum_dy in np.cumsum(dy_hat)]
        x_std = [ math.sqrt(v) for v in np.cumsum([ s**s for s in dy_hat_std])]
        return x, x_std, s


def regress_level_on_first_known(y:Y_TYPE, s:dict, k, a:A_TYPE=None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """ Very basic online regression skater, mostly for testing
           - Only one known in advance variable is utilized
           - Last value is ignored, unless a is None in which case we return 0.0
           - Empirical std is returned
    """
    y0 = wrap(y)[0]  # Ignore contemporaneous, exogenous variables
    if a:
        a0 = wrap(a)[0]  # Ignore all but the first known-in-advance variable

    if not s.get('k'):
        # First invocation
        s = {'p': {}} # Prediction parade
        s['r'] = {}   # Regression state, not to be confused with hyper-param r
        s['k'] = k
        s['o'] = {}   # The "observance" will quarantine 'a' until it can be matched
    else:
        assert s['k']==k  # Immutability

    if a is None:
        return [0]*k, [1.0]*k, s
    else:
        a_t, s['o'] = observance( y=[y0],o=s['o'], k=k, a= [a0])  # Update the observance
        if a_t is not None: # This is the contemporaneous 'a', which was supplied k calls ago.
            if not s['r']:
                # When first calling the online regression algorithm we avoid the degenerate case
                # by sending it two observations.
                y_noise = 0.1*(1e-6+abs(y0))*np.random.randn()
                x_noise = 0.1*(1e-6+abs(a0))*np.random.randn()
                x = [ a_t[0]-x_noise, a_t[0]+x_noise  ]
                y = [ y0-y_noise,  y0+y_noise  ]
                s['r'] = regress_one_helper(x=x, y=y, r=s['r'])
            else:
                s['r'] = regress_one_helper(x=a_t, y=[y0], r=s['r'])

            # Predict using contemporaneous alpha's
            x = [ s['r']['alpha'] + s['r']['beta']*ak[0] for ak in s['o']['a'] ]

            # Push prediction into the parade and get the current bias/stderr
            bias, x_std, s['p'] = parade(p=s['p'], x=x, y=y0)
            return x, x_std, s    # TODO: Use the std implied by regression instead
        else:
            x = [y0]*k
            bias, x_std, s['p'] = parade(p=s['p'], x=x, y=y0)
            return x , x_std, s


LINEAR_SKATERS = [regress_change_on_first_known]


def regress_one_helper(x: [float], y: [float], r:dict)->dict:
    """
        Helper providing online regression for scalar y and a single regressor x.
        Updates a batch of observations.
        From https://stackoverflow.com/questions/52070293/efficient-online-linear-regression-algorithm-in-python

        r - Regression state

                x_avg: average of previous x, if no previous sample, set to 0
                y_avg: average of previous y, if no previous sample, set to 0
                Sxy: covariance of previous x and y, if no previous sample, set to 0
                Sx: variance of previous x, if no previous sample, set to 0
                n: number of previous samples
                new_x: new incoming 1-D numpy array x
                new_y: new incoming 1-D numpy array x
    """
    if not r.get('n_obs'):
        r = {'n_obs': 0,
             'x_avg': 0,
             'y_avg': 0,
             'Sxy': 0,
             'Sx': 0}

    n, x_avg, y_avg, Sxy, Sx = r['n_obs'], r['x_avg'], r['y_avg'], r['Sxy'], r['Sx']

    new_n = n + len(x)
    new_x_avg = (x_avg * n + np.sum(x)) / new_n
    new_y_avg = (y_avg * n + np.sum(y)) / new_n
    if n > 0:
        x_star = (x_avg * np.sqrt(n) + new_x_avg * np.sqrt(new_n)) / (np.sqrt(n) + np.sqrt(new_n))
        y_star = (y_avg * np.sqrt(n) + new_y_avg * np.sqrt(new_n)) / (np.sqrt(n) + np.sqrt(new_n))
    elif n == 0:
        x_star = new_x_avg
        y_star = new_y_avg
    else:
        raise ValueError
    new_Sx = Sx + np.sum((x - x_star) ** 2)
    new_Sxy = Sxy + np.sum((x - x_star).reshape(-1) * (y - y_star).reshape(-1))
    beta = new_Sxy / new_Sx
    alpha = new_y_avg - beta * new_x_avg

    s_posterior = {'Sxy':new_Sxy,'Sx':new_Sx,'n_obs':new_n,'x_avg':new_x_avg,'y_avg':new_y_avg,
                   'alpha':alpha,'beta':beta}

    return s_posterior



if __name__=='__main__':
    # A little test...
    true_beta = 2.0
    a = [ i % 5 for i in range(500) ]
    y = [ true_beta*aj+np.random.randn() for aj in a ]
    s, k = {}, 3
    for aj, yj in zip(a[k:],y):
        x, x_std, s = regress_level_on_first_known(y=yj, s=s, k=k, a=aj)
    assert abs(s['r']['beta']-true_beta)<0.1, 'You broke it somehow! '







from timemachines.skaters.allskaters import SKATERS, skater_from_name, pypi_from_name  # Only those with no hyper-params
from timemachines.skatertools.evaluation.evaluators import evaluate_mean_squared_error_with_sporadic_fit, evaluator_from_name
import numpy as np
from timemachines.skatertools.comparison.eloformulas import elo_update
import time

SLOW_SKATER_KEYWORDS = ['divine','pmd','prophet']

try:
    from microprediction import MicroReader

    have_micro = True
except ImportError:
    have_micro = False

if have_micro:
    from timemachines.skatertools.data.live import random_regular_data as DEFAULT_DATA_SOURCE
else:
    DEFAULT_DATA_SOURCE = None

SKATER_F_FACTOR = 1000  # The scale factor for ratings. In chess this is set to 400.
# If the matchup is considered more fluky than a single game of chess, a higher value might make sense.
SKATER_K_FACTOR = 60  # The Elo update factor (maximum rating gain)


def skater_elo_update(elo: dict, k, evaluator=None, n_burn=400, tol=0.01, initial_elo=1600, data_source=None):
    """ Create or update elo ratings by performing a random matchup on univariate live data

          elo - Dictionary containing the 'state' (i.e. elo ratings and game counts)
          k   - Number of steps to look ahead
          tol - Error ratio that results in a tie being declared
          data_provider - A function taking n_obs and returning y, t

        Speed is *not* taken into account, yet.
    """
    if data_source is None:
        data_source = DEFAULT_DATA_SOURCE
        assert data_source is not None, "If microprediction is not installed you must supply a different data_source function that returns y, t "

    if not elo:
        # Initialize game counts and Elo ratings
        elo['name'] = [f.__name__ for f in SKATERS]
        elo['count'] = [0 for _ in SKATERS]
        elo['rating'] = [initial_elo for _ in SKATERS]
        elo['traceback'] = ['not yet run' for _ in SKATERS]
        elo['active'] = [True for _ in SKATERS]

    else:
        # New fields ... one-off fix
        if not 'pypi' in elo:
            elo['pypi'] = [ pypi_from_name(nm) for nm in elo['name']]
        if not 'seconds' in elo:
            elo['seconds'] = [-1 for nm in elo['name']]

        # Check for newcomers
        new_names = [f.__name__ for f in SKATERS if f.__name__ not in elo['name']]
        for new_name in new_names:
            elo['name'].append(new_name)
            elo['count'].append(0)
            elo['rating'].append(initial_elo)
            elo['traceback'].append('not yet run')
            elo['active'].append(True)
            elo['seconds'].append(-1)
            elo['pypi'].append(pypi_from_name(new_name))

    if evaluator is None:
        if elo.get('evaluator'):
            eval_name = elo['evaluator']
            try:
                evaluator = evaluator_from_name(eval_name)
            except:
                print('Could not retrieve '+eval_name+' so reverting.')
                evaluator = evaluate_mean_squared_error_with_sporadic_fit
        else:
            evaluator = evaluate_mean_squared_error_with_sporadic_fit
        print('Evaluating using '+str(evaluator.__name__))
    elo['evaluator'] = evaluator.__name__

    # Choose two random skaters, but avoid slow ones once
    n_skaters = len(elo['name'])
    i1, i2 = np.random.choice(list(range(n_skaters)), size=2, replace=False)
    skater1, skater2 = elo['name'][i1], elo['name'][i2]
    if any([ slow in skater1 for slow in SLOW_SKATER_KEYWORDS ]) or any([ slow in skater2 for slow in SLOW_SKATER_KEYWORDS ]):
        i1, i2 = np.random.choice(list(range(n_skaters)), size=2, replace=False)
        skater1, skater2 = elo['name'][i1], elo['name'][i2]

    fs = list()
    for i, sn in zip([i1, i2], [skater1, skater2]):
        try:
            f = skater_from_name(sn)
            assert f is not None, 'probably just not used anymore'
            elo['active'][i] = True
            fs.append(f)
        except Exception as e:
            elo['active'][i] = False

    if len(fs) == 2:
        # a pitched paddle battle in a bottle
        print(fs[0].__name__ + ' vs. ' + fs[1].__name__)
        y, t = data_source(n_obs=n_burn + 50)
        scores = list()
        for i, f in zip([i1, i2], fs):
            import traceback
            try:
                st = time.time()
                score = evaluator(f=f, y=y, k=k, a=None, t=t, e_fit=15, e_nofit=-1, n_test=50, fit_frequency=100)
                elo['seconds'][i] = time.time()-st
                elo['traceback'][i] = 'passing'
                scores.append(score)
            except Exception as e:
                elo['traceback'][i] = traceback.format_exc()
                elo['seconds'][i] = st-time.time() # Time taken to fail, as -ve number

        if len(scores) == 2:
            small = tol * (abs(scores[0]) + abs(scores[1]))  # Ties
            points = 1 if scores[0] < scores[1] - small else 0 if scores[1] < scores[0] - small else 0.5
            elo1, elo2 = elo['rating'][i1], elo['rating'][i2]
            min_games = min(elo['count'][i1], elo['count'][i2])
            K = SKATER_K_FACTOR / 2.0 if min_games > 25 else SKATER_K_FACTOR  # The Elo update scaling parameter
            elo['rating'][i1], elo['rating'][i2] = elo_update(elo1, elo2, points, k=K, f=SKATER_F_FACTOR)
            elo['count'][i1] += 1
            elo['count'][i2] += 1
            the_winner = skater1 if points > 0.5 else skater2 if points<0.5 else 'nobody'
            print('     won by '+the_winner)

    return elo



try:
    import pyflux as pf
    using_pyflux = True
except ImportError:
    class Mock:
        ARIMA = None
    pf = Mock()
    print('pyflux is off the list while the package on PyPI is broken. You can install from git')
    print('pip install git+https://github.com/RJT1990/pyflux.git')
    using_pyflux = False


from momentum import rvar  # Could easily add skew, kurtosis
from typing import Union, SupportsFloat, List


# A "parade" is a procession of l-step ahead predictions that are waiting to be judged when data arrives.
# The predictions are stored in an array where the n'th entry will be judged after n data points arrive.
# There are multiple entries in each row of the parade. For example if data is sampled on the hour,
# then a 3-hour ahead forecast made at 11:01 am will march alongside a 5-hour ahead prediction made two hours
# earlier at 9:00am. Both will be judged at 2pm. However they are tagged with the prediction horizon, allowing
# for rolling statistics to be tracked separately for 1-hr, 2-hr,...10-hr ahead predictions, say.
#
# The usage is pretty simple. Send target y and posterior predictions (i.e. those using information y) to
# parade_update, then  use parade_mean, parade_std. See tsaconstant or simple.movingaverage for examples.


def parade(p: dict, x: Union[List[SupportsFloat], None], y: Union[SupportsFloat, None], rho=0.01):
    """
          A 'parade' holds previous predictions and truths, and can be used to determine a running
          estimate of the empirical errors of the predictions. The usage is

          p=parade({})
          ... do something to create predictions x k-steps ahead when you receive y ...
          p = parade(p=p,x=x,y=y)     # submit the prediction vector and the observation y

    :param p:   state - supply empty dict on first call
    :param y:   incoming observation
    :param x:   term structure of predictions out k steps ahead, made after y received
    :param rho: Recency weighting for the empirical errors
    returns:  mean, std, of model residuals and the posterior state s'

    A special convention allows the caller to reset the empirical moments. Pass x=None and y=None

    """
    # Initialize
    if not p:
        k = len(x)
        p = {'predictions': [[] for _ in range(k)],  # Holds the cavalcade
             'moments': [rvar({}, rho=rho, n=5) for _ in range(k)]}  # Could use kurtosis_init here for more moment
    else:
        assert len(x) == len(p['predictions'])  # 'k' is immutable

    if x is None and y is None:
        # This will "reset" the running moments, but keep the existing store of predictions and observations
        p_mean, p_std = parade_mean(p), parade_std(p)
        p['moments'] = [rvar(rho=rho, n=5) for _ in range(k)]
        return p_mean, p_std, p
    else:
        assessable = p['predictions'].pop(0)
        if assessable:
            for j, xi in assessable:
                p['moments'][j] = rvar(p['moments'][j], y - xi)

        p['predictions'].append(list())
        for j, xj in enumerate(x):
            p['predictions'][j].append((j, xj))

        return parade_mean(p), parade_std(p), p


# Helpers for extracting skater-ready predictions

def parade_bias(p):
    """ Model bias is negative of mean residual """
    return [mj.get('mean') for mj in p['moments']]


def parade_mean(p):
    """ Note the sign, E[y-f()] """
    return [noneneg(mj.get('mean')) for mj in p['moments']]


def parade_std(p):
    return [mj.get('std') for mj in p['moments']]


def noneneg(x):
    return -x if x is not None else None


if __name__ == '__main__':
    from pprint import pprint
    import numpy as np

    y = list(range(100))
    x = [yi - 0.5 + np.random.randn() for yi in y][1:]
    p = {}
    for xi, yi in zip(x, y):
        _, _, p = parade(p=p, x=[xi] * 3, y=yi)
    pprint(p)

    print(parade_mean(p=p))
    print(parade_std(p=p))



import pandas as pd
from timemachines.skatertools.utilities.conventions import wrap
from timemachines.skaters.uberorbit.orbitinclusion import using_orbit, LGTMAP, LGTAggregated, LGTFull


if using_orbit:

    def orbit_lgt_skater_factory(y, s, k, e, seasonality):

        y = wrap(y)

        if not s.get('y'):
            s = {'p': {},  # parade
                 'y': list(),  # historical y
                 'k': k}
        else:
            # Assert immutability of k, dimensions of y,a
            if s['y']:
                assert len(y) == len(s['y'][0])
                assert k == s['k']

        if y is None or (e is not None and e < 0):
            return None, None, s
        else:
            s['y'].append(y)

            if e is None:
                burn_in_period = 410
            else:
                burn_in_period = int(e)

            if len(s['y']) > burn_in_period:
                y_arg = s['y'][k:]
                x, x_std = run_orbit_lgt(s['y'], k, seasonality)
            else:
                x = [y[0]] * k
                x_std = [1] * k

        return x, x_std, s

    def run_orbit_lgt(data,k,seasonality):

        #It seems uberorbit needs a two column dataframe of an index column and
        #measuremnt column in order to work. It also seems to need enough
        #space in the dataframe to insert the values it predicts.
        df = pd.DataFrame(data).reset_index().rename(columns={0:"response_col",'index':"date_col"})
        for i in range(k):
            df2 = pd.DataFrame([[len(df),0]], columns=["date_col","response_col"])
            df = df.append(df2, ignore_index=True)

        #"test/train" split
        test_size = k
        train_df = df[:-test_size]
        test_df = df[-test_size:]

        #run the model
        if isinstance(seasonality, int) or seasonality == None:
            lgt = LGTMAP(
                response_col="response_col",
                date_col="date_col",
                seasonality=seasonality
            )
        else:
            print("The seasonality given was not an int or None type. Try again...")
            return
        lgt.fit(df=train_df)

        #get the prediction(s)
        predicted_df = lgt.predict(df=test_df)

        #list containing predictions 1 through k steps out
        x = list(predicted_df.prediction)
        #standard deviaiton list (this is technically 2 std)
        x_std = [a / 2 for a in list(predicted_df.prediction_95 - predicted_df.prediction_5)]

        return x, x_std


    def orbit_lgt_12(y,s,k,e):
        return orbit_lgt_skater_factory(y, s, k, e, seasonality=12)


    def orbit_lgt_24(y,s,k,e):
        return orbit_lgt_skater_factory(y, s, k, e, seasonality=24)





PMD_CRITERIA = ['aic', 'aicc', 'bic', 'hqic']


def pmd_params(method=None):
    """ Default hyper-params """
    # https://alkaline-ml.com/pmdarima/modules/generated/pmdarima.arima.auto_arima.html
    s = dict(start_p=1, start_q=1, start_P=1, start_Q=1,
                       max_p=10, max_q=10, max_P=24, max_Q=24, seasonal=False,
                       stepwise=True, suppress_warnings=True,
                       D=None, max_D=3,
                       error_action='ignore',
                       information_criterion=PMD_CRITERIA[0])
    if method == '0':
        # Add more cases here
        pass
    return s


from timemachines.skatertools.data.synthetic import brownian_with_noise, brownian_with_exogenous
from timemachines.skating import prior, residuals, prior_with_sporadic_fit
from sklearn.metrics import mean_squared_error, mean_absolute_error
from scipy.stats import energy_distance
import numpy as np
from timemachines.skatertools.utilities.conventions import targets
from typing import List
from timemachines.skatertools.data.real import hospital, hospital_with_exog

# Evaluation of skaters


def evaluator_from_name(name):
    valid = [f for f in EVALUATORS if f.__name__==name ]
    return valid[0] if len(valid)==1 else None


def evaluate_sklearn_metric(f, y, k:int, a=None, t=None, e_fit=60, e_nofit=-1, metric=None, r=None)->float:
    """ Compute prior for skater and evaluate an sklearn metric

           n_test:   Number of data points to test with

        Trains on history then computes several test samples
    """
    return evaluate_sklearn_metric_with_sporadic_fit(f=f,y=y,k=k,a=a,t=t,e_fit=e_fit, e_nofit=e_nofit, r=r, fit_frequency=1, metric=metric )


def evaluate_sklearn_metric_with_sporadic_fit(f, y, k:int, a=None, t=None, e=None, r=None, metric=None,
                                              n_test:int=10, e_fit=60,  e_nofit=-1, fit_frequency:int=100)->float:
    x, x_std = prior_with_sporadic_fit(f=f, y=y, k=k, a=a, t=t, r=r, n_test=n_test,
                                       e_fit=e_fit,  e_nofit=e_nofit, fit_frequency=fit_frequency)
    yt = targets(y)
    xk = [xt[-1] for xt in x]
    return metric(yt[-n_test:], xk[-n_test:])


def evaluate_mean_squared_error_with_sporadic_fit(f, y, k:int, a=None, t=None, r=None, n_test:int=10, e_fit=60, e_nofit=-1, fit_frequency:int=100)->float:
    return evaluate_sklearn_metric_with_sporadic_fit(f=f, y=y, k=k, a=a, t=t, r=r, metric=mean_squared_error, e_fit=e_fit, e_nofit=e_nofit, n_test=n_test, fit_frequency=fit_frequency)


def hospital_mean_square_error_with_sporadic_fit(f, k:int=1, n=120, r=None, n_test:int=10, e_fit=60, e_nofit=-1, fit_frequency:int=100)->float:
    """ Useful for a quick test of a skater, univariate and random hyper-param """
    y = hospital()[:n]
    return evaluate_mean_squared_error_with_sporadic_fit(f=f, y=y, k=k, r=r, e_fit=e_fit, n_test=n_test, e_nofit=e_nofit, fit_frequency=fit_frequency)


def hospital_exog_mean_square_error_with_sporadic_fit(f, k, n=120, r=None, n_test:int=10, e_fit=60, e_nofit=-1, fit_frequency:int=100)->float:
    """ Useful for a quick test of a skater w/ exogenous inputs and known-in-advance variables """
    y, a = hospital_with_exog(n=n,k=k)
    return evaluate_mean_squared_error_with_sporadic_fit(f=f, y=y, a=a, k=k, r=r, n_test=n_test, e_fit=e_fit, e_nofit=e_nofit, fit_frequency=fit_frequency)


# Energy distance between consecutive epochs
# (a more speculative way to evaluate point estimates)


def chunk_to_end(l:List, n:int)-> List[List]:
    """ Break list in to evenly sized chunks
        :param n: Size of batches
    """
    rl = list(reversed(l))
    chunks = [ list(reversed(rl[x:x + n])) for x in range(0, len(rl), n) ]
    return list(reversed(chunks[:-1]))


def evaluate_energy(f, y=None, k=1, a=None, t=None, e=None, r=None, n_burn=30, n_epoch=30):
    r = residuals(f=f, y=y, k=k, a=a, t=t, e=e, r=r, n_burn=n_burn)
    r_chunks = chunk_to_end(r,n_epoch)
    assert len(r_chunks)>=2,'Cannot evaluate ... try a shorter n_epoch '
    uv = [ (u,v) for u,v in zip( r_chunks[1:],r_chunks[:-1] )]   # residuals in one epoch versus the next
    return np.mean([ energy_distance(u_values=u_values,v_values=v_values) for u_values, v_values in uv ])


def hospital_energy(f, k=3, n=100, n_epoch=20, n_burn=18):
    y, a = hospital_with_exog(n=n+5)
    return evaluate_energy(f=f, y=y, k=k, a=a, n_burn=n_burn, n_epoch=n_epoch)


def brownian_energy(f, n=500, **kwargs):
    ys = brownian_with_noise(n)
    return evaluate_energy(f=f, y=ys, **kwargs)


def exogenous_energy(f, n=500, **kwargs):
    ys = brownian_with_exogenous(n)
    return evaluate_energy(f=f, y=ys, **kwargs)


EVALUATORS = [evaluate_mean_squared_error_with_sporadic_fit]


from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any
from timemachines.skatertools.components.parade import parade
from timemachines.skaters.proph.prophiskaterfactory import using_prophet
if using_prophet:
    from timemachines.skaters.proph.prophiskaterfactory import prophet_iskater_factory
from timemachines.skatertools.utilities.nonemath import nonecenter
from timemachines.skaters.proph.prophparams import PROPHET_META, prophet_params
from timemachines.skatertools.utilities.nonemath import nonecast
import sys
import logging
import numpy as np

logging.disable(sys.maxsize)
logging.getLogger('fbprophet').setLevel(logging.ERROR)


###################################################################################################
#                                                                                                 #
#                      Facebook Prophet skater factory                                            #
#                                                                                                 #
# Wraps prophet model with recursive use, running empirical moments, and some hyper-parameters    #
# See https://facebook.github.io/prophet/docs/diagnostics.html#hyperparameter-tuning              #
#                                                                                                 #
# Advantages:                                                                                     #
#       - State is a simple dictionary, since prophet itself has no notion of state               #
#       - Supposed to work okay without any hyper-param tuning                                    #
#       - Great documentation, response to GitHub issues and backing by Facebook devs             #
#                                                                                                 #
# Disadvantages:                                                                                  #
#       - Model requires re-fitting after each and every data point                               #
#       - Generative model isn't suggestive of strong out of sample performance (opinion)         #
#       - Empirical results are quizzical                                                         #
#                                                                                                 #
# See also: prophetiskaterfatory                                                                  #                              #
#                                                                                                 #
###################################################################################################

if using_prophet:

    def fbprophet_skater_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None,
                                 emp_mass: float = 0.0, emp_std_mass: float = 0.0,
                                 freq=None, recursive: bool = False,
                                 model_params: dict = None,
                                 n_max: int = None) -> ([float], Any, Any):
        """ Prophet skater with running prediction error moments
            Hyper-parameters are explicit here, whereas they are determined from r in actual skaters.
            Params of note:

                 a: value of known-in-advance vars k step in advance (not contemporaneous with y)

        """

        assert 0 <= emp_mass <= 1
        assert 0 <= emp_std_mass <= 1
        if e is None:
          e = 1  # 'Providing e=None to prophet is not recommended. Use e<0 to skip a fitting or e>0 to fit.'

        if freq is None:
            freq = PROPHET_META['freq']
        if n_max is None:
            n_max = PROPHET_META['n_max']

        y = wrap(y)
        a = wrap(a)

        if not s.get('y'):
            s = {'p': {},     # parade
                 'y': list(), # historical y
                 'a': list(), # list of a known k steps in advance
                 't': list(),
                 'k': k}
        else:
            # Assert immutability of k, dimensions of y,a
            if s['y']:
                assert len(y) == len(s['y'][0])
                assert k == s['k']
            if s['a']:
                assert len(a) == len(s['a'][0])

        if y is None:
            return None, s, None
        else:
            s['y'].append(y)
            if a is not None:
                s['a'].append(a)
            if t is not None:
                assert isinstance(t,float), 'epoch time please'
                s['t'].append(t)

            if e is not None and (e>0) and (len(s['a'])>k) and (len(s['y'])>10):
                # Offset y, t, a are supplied to prophet interface
                # Prophet requires several non-nan rows
                t_arg = s['t'][k:] if t is not None else None
                a_arg = s['a']
                y_arg = s['y'][k:]
                x, x_std, forecast, model = prophet_iskater_factory(y=y_arg, k=k, a=a_arg, t=t_arg,
                                                                    freq=freq, n_max=n_max,
                                                                    recursive=recursive, model_params=model_params)
                s['m'] = True # Flag indicating a model has been fit (there is no point keeping the model itself, however)
            else:
                x = [y[0]] * k
                x_std = [1] * k

            # Get running mean prediction errors from the prediction parade
            x_resid, x_resid_std, s['p'] = parade(p=s['p'], x=x, y=y[0])
            x_resid = nonecast(x_resid,y[0])
            x_resid_std = nonecast(x_resid_std,1.0)

            # Compute center of mass between bias-corrected and uncorrected predictions
            x_corrected = np.array(x_resid) + np.array(x)
            x_center = nonecenter(m=[emp_mass, 1 - emp_mass], x=[x_corrected, x])
            x_std_center = nonecenter(m=[emp_std_mass, 1 - emp_std_mass], x=[x_resid_std, x_std])

            return x_center, x_std_center, s


    def fbprophet_hyperparam_skater_factory(r: R_TYPE = None, param_names: [str] = None,   **kwargs):
        """ Useful for creating skaters based on hyper-parameters r and the
            method of modifying them suggested by the authors
         """
        assert param_names is not None
        dim = len(param_names)
        assert 2 <= dim <= 3
        model_params = prophet_params(r=r,dim=dim, param_names=param_names)
        return fbprophet_skater_factory(model_params=model_params, **kwargs)


    def fbprophet_skater_testor(y :Y_TYPE, s:dict=None, k:int =1, a:A_TYPE =None,
                         t:T_TYPE=None, e:E_TYPE =None, r:R_TYPE =None, freq=None, n_max=None):
        """ A default facebook prophet usage, with no hyper-parameters and no prediction parade """
        # For testing

        if freq is None:
            freq = PROPHET_META['freq']
        if n_max is None:
            n_max = PROPHET_META['n_max']

        y = wrap(y)
        a = wrap(a)

        if not s.get('y'):
            s = {'y': list(),
                 'a': list(),
                 'k': k}
        else:
            # Assert immutability of k, dimensions
            if s['y']:
                assert len(y) == len(s['y'][0])
                assert k == s['k']
            if s['a']:
                assert len(a) == len(s['a'][0])

        if y is None:
            return None, s, None
        else:
            s['y'].append(y)
            if a is not None:
                s['a'].append(a)
            if len(s['y']) > max(2*k+5,PROPHET_META['n_warm']):
                x, x_std, _, _ = prophet_iskater_factory(y=s['y'], k=k, a=s['a'], freq=freq, n_max=n_max)
            else:
                x = [y[0]] * k
                x_std = [1.0] * k
            return x, x_std, s


if __name__ == '__main__':
    from timemachines.skatertools.data import hospital_with_exog
    from timemachines.skatertools.evaluation.evaluators import evaluate_mean_absolute_error

    k = 3
    y, a = hospital_with_exog(k=k, n=100, offset=True)
    f = fbprophet_skater_factory
    err2 = evaluate_mean_absolute_error(f=f, k=k, y=y, a=a, n_burn=50)
    print(err2)



try:
    from neuralprophet import NeuralProphet
    using_neuralprophet = True
except ImportError:
    class NeuralMock:
        whatever=False
    using_neuralprophet = False
    NeuralProphet = NeuralMock()


try:
    from microprediction import MicroReader
    have_micro = True
except ImportError:
    have_micro = False

import random

# Live data streams at www.microprediction.com

if have_micro:

    def random_stream_name(min_len=500,exclude_str=None, include_str=None):
        """ Randomly selected univariate series
                :return:  y, t, url
                """
        mr = MicroReader()
        names = mr.get_stream_names()
        okay = False
        while not okay:
            name = random.choice(names)
            n_obs = len(mr.get_lagged_values(name=name, count=10000))
            okay = True
            if exclude_str is not None and exclude_str in name:
                okay = False
            if include_str is not None and include_str not in name:
                okay = False
            if n_obs < min_len:
                okay = False
        url = 'https://www.microprediction.org/stream_dashboard.html?stream=' + name.replace('.json', '')
        return name, url

    def random_regular(min_len=500):
        """ Randomly selected univariate series
        :return:  y, t, url
        """
        return random_stream_name(min_len=min_len,exclude_str='~')


    def random_residual(min_len=500):
        """ Randomly selected univariate series
        :return:  y, t, url
        """
        return random_stream_name(min_len=min_len, include_str='z1~')


    def stream_data(name:str,n_obs:int):
        """ values and times for a univariate stream """
        mr = MicroReader()
        lagged_values, lagged_times = mr.get_lagged_values_and_times(name=name, count=n_obs)
        y, t = list(reversed(lagged_values)), list(reversed(lagged_times))
        return y, t


    def random_regular_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
        :returns:  y, t
        """
        name, url = random_regular(min_len=n_obs)
        return stream_data(name=name,n_obs=n_obs)


    def random_residual_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
        :returns:  y, t
        """
        name, url = random_residual(min_len=n_obs)
        return stream_data(name=name, n_obs=n_obs)




if __name__=='__main__':
    if have_micro:
        name, url = random_regular(min_len=900)
        print(url)


from timemachines.skatertools.ensembling.ensemblefactory import precision_weighted_ensemble_factory, R_BALANCED, R_AGGRESSIVE, R_PRECISION
from timemachines.skaters.tsa.tsaconstant import TSA_D0_SKATERS
from timemachines.skatertools.utilities.conventions import Y_TYPE,A_TYPE,T_TYPE,E_TYPE
from typing import Any


def tsa_precision_d0_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    return precision_weighted_ensemble_factory(fs=TSA_D0_SKATERS,y=y,s=s,k=k,a=a,t=t,e=e,r=R_PRECISION)


def tsa_balanced_d0_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    return precision_weighted_ensemble_factory(fs=TSA_D0_SKATERS,y=y,s=s,k=k,a=a,t=t,e=e,r=R_BALANCED)


def tsa_aggressive_d0_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    return precision_weighted_ensemble_factory(fs=TSA_D0_SKATERS,y=y,s=s,k=k,a=a,t=t,e=e,r=R_AGGRESSIVE)

TSA_DO_ENSEMBLE_SKATERS  = [tsa_precision_d0_ensemble, tsa_balanced_d0_ensemble, tsa_aggressive_d0_ensemble]






def residual(s:dict, y, x):
    """ Return residuals
      :param s:   state - supply empty dict on first call
      :param y:   incoming observation
      :param x:   term structure of predictions out k steps ahead, made after y received
      :returns k-vector of residuals
    """
    k = len(x)
    if not s:
        s = {'predictions': [[] for _ in range(k)]}  # Holds the cavalcade
    else:
        assert len(x) == len(s['predictions'])  # 'k' is immutable

    assessable = s['predictions'].pop(0)
    z = [0 for _ in range(k)]
    if assessable:
        for j, xi in assessable:
            z[j] = y - xi

    s['predictions'].append(list())
    for j, xj in enumerate(x):
        s['predictions'][j].append((j, xj))

    return z, s


from timemachines.skaters.proph.prophiskaterfactory import using_prophet
if using_prophet:
    from timemachines.skaters.proph.prophskaterfactory import fbprophet_skater_factory, fbprophet_hyperparam_skater_factory
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
    import numpy as np

    # A collection of skaters powered by fbprophet
    # You'll want to read this review: https://www.microprediction.com/blog/prophet


    def fbprophet_exogenous(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Uses known-in-advance and also y[1:] brought forward """
        return fbprophet_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e)


    def fbprophet_recursive(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Same as exogenous, but uses prophet to predict y[1:]  """
        return fbprophet_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, recursive=True)


    def fbprophet_known(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Uses known-in-advance but not y[1:] """
        y0 = [wrap(y)[0]]
        return fbprophet_skater_factory(y=y0, s=s, k=k, a=a, t=t, e=e)


    def fbprophet_univariate(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Simple univariate prediction using only y[0], and not 'a' or y[1:] """
        y0 = [wrap(y)[0]]
        return fbprophet_skater_factory(y=y0, s=s, k=k, a=None, t=t, e=e)



    def fbprophet_cautious(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Similar to fbexogenous, but no crazy nonsense """
        if not s.get('s'):
            s['s']={}       # prophet's state
            s['y']=list()   # maintain last five values
        y0 = wrap(y)[0]
        s['y'].append(y0)
        if len(s['y'])>5:
            s['y'].pop(0)
        import math
        x_upper = [ np.max(s['y'])+math.sqrt(j+1)*np.std(s['y']) for j in range(k) ]
        x_lower = [ np.min(s['y'])-math.sqrt(j+1)*np.std(s['y']) for j in range(k) ]
        x, x_std, s['s'] = fbprophet_univariate(y=y,s=s['s'],k=k,a=a,t=t,e=e)
        x_careful = np.minimum(np.array(x),np.array(x_upper))
        x_careful = np.maximum(x_careful, np.array(x_lower))
        return list(x_careful), x_std, s


    PROPHET_SKATERS_SINGULAR = [fbprophet_exogenous, fbprophet_known, fbprophet_univariate, fbprophet_recursive,
                                fbprophet_cautious]


    # (1) Skaters with author-suggested two-dimensional hyper-parameter spaces

    def fbprophet_exogenous_r2(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r:R_TYPE=None):
        """ A skater using exogenous variables, with hyper-param tuning as recommended by authors """
        assert r is not None
        param_names = ['changepoint_prior_scale', 'seasonality_prior_scale']
        return fbprophet_hyperparam_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, param_names=param_names)


    def fbprophet_recursive_r2(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r:R_TYPE=None):
        """ Same as exogenous, but uses prophet to predict y[1:]  """
        assert r is not None
        param_names = ['changepoint_prior_scale', 'seasonality_prior_scale']
        return fbprophet_hyperparam_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, param_names=param_names, recursive=True)


    def fbprophet_known_r2(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r:R_TYPE=None):
        """ Uses known-in-advance but not y[1:] """
        assert r is not None
        y0 = [wrap(y)[0]]
        param_names = ['changepoint_prior_scale', 'seasonality_prior_scale']
        return fbprophet_hyperparam_skater_factory(y=y0, s=s, k=k, a=a, t=t, e=e, r=r, param_names=param_names, recursive=False)


    def fbprophet_univariate_r2(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None,r:R_TYPE=None):
        """ Simple univariate prediction using only y[0], and not 'a' or y[1:] """
        assert r is not None
        y0 = [wrap(y)[0]]
        param_names = ['changepoint_prior_scale', 'seasonality_prior_scale']
        return fbprophet_hyperparam_skater_factory(y=y0, s=s, k=k, a=None, t=t, e=e, r=r, param_names=param_names, recursive=False)


    PROPHET_R2_SKATERS = [ fbprophet_exogenous_r2, fbprophet_known_r2, fbprophet_univariate_r2, fbprophet_recursive_r2 ]
else:
    PROPHET_SKATERS_SINGULAR = []
    PROPHET_R2_SKATERS = []



if __name__ == '__main__':
    from timemachines.skatertools.data import hospital_with_exog
    from timemachines.skatertools.visualization.priorplot import prior_plot
    import matplotlib.pyplot as plt
    k = 1
    y, a = hospital_with_exog(k=k, n=450, offset=True)
    f = fbprophet_exogenous
    err2 = prior_plot(f=f, k=k, y=y, n=450, n_plot=50)
    print(err2)
    plt.show()
    pass





from timemachines.skaters.nproph.nprophetinclusion import using_neuralprophet, NeuralProphet
if using_neuralprophet:
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE
    from timemachines.skatertools.composition.residualshypocratic import quickly_moving_hypocratic_residual_factory
    from timemachines.skaters.nproph.nprophetskaters import nprophet_p1, nprophet_p2, nprophet_p3, nprophet_p5, nprophet_p8


    def nprophet_p1_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Chase residuals, somewhat cautiously using, quickly moving average """
        return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=None, f=nprophet_p1)


    def nprophet_p2_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Chase residuals, somewhat cautiously using, quickly moving average """
        return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=None, f=nprophet_p2)


    def nprophet_p3_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Chase residuals, somewhat cautiously using, quickly moving average """
        return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=None, f=nprophet_p3)


    def nprophet_p5_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Chase residuals, somewhat cautiously using, quickly moving average """
        return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=None, f=nprophet_p5)


    def nprophet_p8_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Chase residuals, somewhat cautiously using, quickly moving average """
        return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=None, f=nprophet_p8)

    NPROPHET_SKATERS_COMPOSED = [nprophet_p1_hypocratic, nprophet_p2_hypocratic, nprophet_p3_hypocratic,
                                 nprophet_p5_hypocratic, nprophet_p8_hypocratic ]
else:
    NPROPHET_SKATERS_COMPOSED = []



if __name__ == '__main__':
    assert using_neuralprophet,'pip install neuralprophet'
    from timemachines.skatertools.data.real import hospital_with_exog
    from timemachines.skatertools.visualization.priorplot import prior_plot
    import matplotlib.pyplot as plt
    k = 1
    y, a = hospital_with_exog(k=k, n=450, offset=True)
    f = nprophet_p1_hypocratic
    err2 = prior_plot(f=f, k=k, y=y, n=450, n_plot=50)
    print(err2)
    plt.show()
    pass




try:
    from pydlm import dlm, trend, seasonality, dynamic, autoReg
    using_dlm = True
except ImportError:
    class Nothing:
        yup = False
    using_dlm = False
    dlm=Nothing()
    trend=Nothing()
    seasonality=Nothing()
    dynamic =Nothing()
    autoReg=Nothing()


from timemachines.skaters.dlm.dlmunivariate import using_dlm
if using_dlm:
    from timemachines.skaters.dlm.dlmunivariate import dlm_univariate_a, dlm_univariate_b
    from timemachines.skatertools.composition.residualshypocratic import quickly_moving_hypocratic_residual_factory
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE


    def dlm_univariate_hypocratic_a(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        return quickly_moving_hypocratic_residual_factory(f=dlm_univariate_a, y=y,s=s,k=k,a=a,t=t,e=e)


    def dlm_univariate_hypocratic_b(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        return quickly_moving_hypocratic_residual_factory(f=dlm_univariate_b, y=y,s=s,k=k,a=a,t=t,e=e)


    DLM_SKATERS_COMPOSED = [ dlm_univariate_a, dlm_univariate_b ]
else:
    DLM_SKATERS_COMPOSED = []





from typing import Callable
from timemachines.skatertools.utilities.conventions import Y_TYPE, K_TYPE, A_TYPE, T_TYPE, E_TYPE,\
    E_TYPE_OR_LIST, R_TYPE, targets
import numpy as np
import math


###########################################################################
#                                                                         #
#  Sequential assimilation of observations using a skater, strongly       #
#  suggesting what a skater function is supposed to accomplish, and how   #
#  it should behave                                                       #
#                                                                         #
#  See also skatertools.evaluation.evaluators                             #
#           skatertools.visualization.priorplot.prior_plot                #
#                                                                         #
###########################################################################
from typing import Union


def posterior(f:Callable,       # "Model" to run forward through observations
              y:[Y_TYPE],       # Observations
              k:K_TYPE=1,       # Steps ahead to predict, typically integer
              a:[A_TYPE]=None,  # Data known in advance, or maybe action-conditional variables
              t:[T_TYPE]=None,  # Times of observations (epoch seconds)
              e:E_TYPE_OR_LIST=None,  # Computation time limit per observation during 'fast' stage
              r:R_TYPE=None):   # Optional, will alter broadcast of e, if e is scalar or None

    """ Sequentially feed all data to skater and accumulate predictions

         returns:  x:     [ [float] ]    list of all prediction vectors
                   x_std: [ [float] ]    list of all std vectors
    """

    # Broadcasting
    a = [None for _ in y] if a is None else a
    t = [None for _ in y] if t is None else t
    e = [None for _ in y] if e is None else e

    # Hyper-param must be in [0,1]
    if r is not None:
        assert 0 <= r <= 1

    # Feed one data point at a time to the skater ...
    s = dict()  # <- Skater will receive empty dictionary on the first invocation
    x = list()
    x_std = list()
    for yi, ai, ti, ei in list(zip(y, a, t, e)):
        if r is None:
            xi, x_stdi, s = f(y=yi, s=s, k=k, a=ai, t=ti, e=ei)
        else:
            xi, x_stdi, s = f(y=yi, s=s, k=k, a=ai, t=ti, e=ei, r=r)
        x.append(xi)
        x_std.append(x_stdi)
    return x, x_std


def prior(f, y:[Y_TYPE], k:int, a:[A_TYPE]=None, t:[T_TYPE]=None, e:E_TYPE_OR_LIST=None, r:float=None,
          x0:float=np.nan, std0:float=np.nan):
    """ Compute k-step ahead estimates then align predictions with the observations y, for
        easy error computation. Pre-pad with naive forecast. Ignore last k data points.
    """
    x_post, x_std = posterior(f=f, y=y[:-k], k=k, a=a, t=t, e=e, r=r)
    x_prior = [[x0]*k]*k + list(x_post)
    x_prior_std = [[std0]*k]*k + list(x_post)
    return x_prior, x_prior_std


def prior_with_sporadic_fit(f, y:[Y_TYPE], k:int, a=None, t=None, e_fit=60,  e_nofit=-1, fit_frequency:int=100, r=0.5, x0=np.nan, n_test:int=10):
    """ Similar to the above, but we send a pattern like
         e = [ -1 .. -1 60 -1 .. -1 60 -1 ... -1, 60 60 60 60 60 60 60 ] to reduce computation time during a burn-in period

           e_fit - Value of e to provide when fit is desired (written as E above)
           train_frequency - How often to suggest fit during burn-in period
           n_test - The number of data points for which the skater is instructed to "do something that might take a while" (i.e. e>0 )

    """
    n_reps = math.ceil(len(y) / fit_frequency)
    e = (([e_nofit] * (fit_frequency - 1) + [e_fit]) * n_reps)[:len(y) - k - n_test] + [e_fit] * n_test
    return prior(f=f, y=y, k=k, a=a, t=t, e=e, r=r, x0=x0)


def residuals(f, y:[Y_TYPE], k:int, a=None, t=None, e=None, r=0.5, x0=np.nan, n_burn=50):
    """ Feed fast skater all data points, then report residuals """
    assert n_burn>k
    x, _ = prior(f=f, y=y, k=k, a=a, t=t, e=e, r=r, x0=x0)
    yt = targets(y)
    xk = [ xt[-1] for xt in x]
    return np.array(xk[-n_burn:])-np.array(yt[-n_burn:])




from timemachines.skaters.divine.divineinclusion import using_divinity, dv
if using_divinity:
    import numpy as np
    from timemachines.skatertools.utilities.conventions import Y_TYPE, K_TYPE
    from timemachines.skatertools.utilities.suppression import no_stdout_stderr
    from timemachines.skatertools.visualization.priorplot import prior_plot
    from timemachines.skatertools.utilities.conventions import wrap
    from timemachines.skatertools.components.parade import parade
    from timemachines.skatertools.utilities.nonemath import nonecast
    from timemachines.skaters.divine.divineparams import MIN_N_WARM, DIVINE_MODEL
    from copy import deepcopy

    # This guy is a little chatty

    def divinity_univariate_factory(y:Y_TYPE, s, k:K_TYPE, a=None, t=None, e=None,
                                    max_buffer_len=1000,
                                    n_warm = 101,
                                    model_params:dict=None):
        """ A partial wrapping of the divinity library with notable limitations:

             - Fits every invocation
             - Ignores exogenous variables
             - State is merely a buffer

        """
        y0 = wrap(y)[0]
        assert n_warm>=101,' You must use n_warm'

        if not s:
            s = dict(y=[])

        if not s.get('p'):
            s['p'] = {}   # Initialize prediction parade

        if y0 is None:
            return None, None, s   # Ignore suggestion to fit offline

        # Update buffer
        s['y'].append(y0)
        if len(s['y'])>max_buffer_len+1000:
            s['y'] = s['y'][-max_buffer_len:]

        # Fit and predict, if warm, or just last value
        if len(s['y']) < max( n_warm, MIN_N_WARM ):
            return [y0]*k, [abs(y0)]*k, s
        else:
            with no_stdout_stderr():
                kwargs = deepcopy(DIVINE_MODEL)
                if model_params:
                    kwargs.update(**model_params)
                model = dv.divinity(forecast_length=k,**kwargs)
                model.fit(np.array(s['y']))
            x = list(model.predict())
            _we_ignore_bias, x_std, s['p'] = parade(p=s['p'], x=x, y=y0)
            x_std_fallback = nonecast(x_std, fill_value=1.0)
            return x, x_std_fallback, s


if __name__=='__main__':
    assert using_divinity,'pip install divinity'
    if using_divinity:
        err = prior_plot(f=divinity_univariate_factory, k=5, n=500)
        pass
    else:
        print('pip install divinity')




from timemachines.skaters.simple.movingaverage import EMA_SKATERS, EMA_R1_SKATERS
from timemachines.skaters.simple.linear import LINEAR_SKATERS
from timemachines.skaters.divine.divineskaters import DIVINE_SKATERS
from timemachines.skaters.proph.allprophetskaters import PROPHET_SKATERS, PROPHET_R2_SKATERS
from timemachines.skaters.dlm.alldlmskaters import DLM_SKATERS
from timemachines.skaters.simple.thinking import THINKING_SKATERS
from timemachines.skaters.simple.hypocraticensemble import HYPOCRATIC_ENSEMBLE_SKATERS
from timemachines.skaters.pmd.allpmdskaters import PMD_SKATERS
from timemachines.skaters.tsa.alltsaskaters import TSA_SKATERS
from timemachines.skaters.nproph.allnprophetskaters import NPROPHET_SKATERS
from timemachines.skaters.uberorbit.allorbitskaters import ORBIT_SKATERS

SKATERS_R3 = []
SKATERS_R2 = PROPHET_R2_SKATERS
SKATERS_R1 = EMA_R1_SKATERS

# Leaving out LINEAR_SKATERS

# And with no hyper-params...
SKATERS = EMA_SKATERS + PROPHET_SKATERS + DIVINE_SKATERS + DLM_SKATERS + \
          THINKING_SKATERS + PMD_SKATERS + TSA_SKATERS + NPROPHET_SKATERS + HYPOCRATIC_ENSEMBLE_SKATERS + ORBIT_SKATERS


def skater_from_name(name):
    valid = [f for f in SKATERS if f.__name__==name ]
    return valid[0] if len(valid)==1 else None


PYPI = {'tsa':'statsmodel',
        'fbprophet':'prophet',
        'pmd':'pmdarima',
        'flux':'pyflux',
        'nprophet':'neuralprophet',
        'dlm':'pydlm',
        'divine':'divinity',
        'orbit':'orbit'}


def pypi_from_name(name):
    stem = name.split('_')[0]
    short_name = PYPI.get(stem)
    stub = 'https://pypi.org/project/'
    return stub+short_name if short_name else stub+'timemachines'


if __name__=='__main__':
    from pprint import pprint
    pprint([(sk.__name__,pypi_from_name(sk.__name__)) for sk in SKATERS])
    print(len(SKATERS))


try:
    import orbit
    from orbit.models.lgt import LGTMAP, LGTAggregated, LGTFull
    using_orbit = True
except ImportError:
    LGTMAP = None
    LGTAggregated = None
    LGTFull = None
    using_orbit = False



from timemachines.skaters.dlm.dlminclusion import using_dlm, trend, seasonality, dynamic
if using_dlm:
    from timemachines.skatertools.utilities.conventions import dimension, to_int_log_space, split_exogenous, Y_TYPE, R_TYPE
    from timemachines.skatertools.visualization.priorplot import prior_plot_exogenous
    from timemachines.skaters.dlm.dlmhacks import quietDlm, fixedAutoReg
    from timemachines.skatertools.utilities.conventions import wrap
    import math
    import numpy as np


    def dlm_exogenous_a(y, s: dict, k: int, a=None, t=None, e=None):
        ra = 0.25
        return dlm_exogenous_r3(y=y, s=s, k=k,  a=a, t=t, e=e, r=ra)


    def dlm_exogenous_b(y, s: dict, k: int, a=None, t=None, e=None):
        rb = 0.75
        return dlm_exogenous_r3(y=y, s=s, k=k, a=a, t=t, e=e, r=rb)


    def dlm_exogenous_r3(y, s, k, a, t, e, r):
        """ One way to use dlm
            :returns: x, s', w
        """
        if not s:
            s = dict()
            s['dim'] = dimension(y)
            s = dlm_set_exog_hyperparams(s=s, r=r)
            y0, exog = split_exogenous(y=y)
            s['n_obs'] = 0
            s['model'] = quietDlm([], printInfo=False) + trend(s['trend_degree'], s['discount']) + seasonality(s['period'], s['discount'])
            s['model'] = s['model'] + fixedAutoReg(degree=s['auto_degree'], name='ar', w=1.0)
            if exog:
                exog_wrapped = [[None if np.isnan(ex0) else ex0 for ex0 in exog]]
                s['model'] = s['model'] + dynamic(features=exog_wrapped, discount=0.99, name='exog') # Set's first exog

        if y is not None:
            y = wrap(y)
            assert dimension(y)==s['dim'],'Cannot change dimension of data sent'
            s['n_obs'] += 1
            y0, exog = split_exogenous(y=y)
            y0_passed_in = None if np.isnan(y0) else y0  # pydlm uses None for missing values
            s['model'].append([y0_passed_in])
            if exog:
                exog_wrapped = [[ None if np.isnan(ex0) else ex0 for ex0 in exog ]]
                if s['n_obs']>1:
                    s['model'].append(data=exog_wrapped, component='exog') # Don't get first exog twice
            num_obs = len(s['model'].data) if s.get('model') else 0
            if num_obs % s['n_fit'] == s['n_fit']-1:
                _, _, s = dlm_exogenous_r3(y=None, s=s, k=k, a=a, t=t, e=10, r=r)
            s['model'].fitForwardFilter()
            return _dlm_exog_prediction_helper(s=s, k=k, y=y)

        if y is None:
            if dimension(y)==1:
                s['model'].tune(maxit=20)
                # Don't tune if exogenous ... haven't got this to work
            s['model'].fit()
            return None, None, s



    def _dlm_exog_prediction_helper(s, k:int, y:Y_TYPE):
        num_obs = len(s['model'].data) if s.get('model') else 0
        if num_obs < s['n_burn']:
            y0, exog = split_exogenous(y)
            return [y0]*k, [abs(y0)]*k, s
        else:
            assert k==1,'only k==1 for now' # TODO: Fix to allow for k>1
            y0, exog = split_exogenous(y)
            if exog:
                exog_passed_in = [None if np.isnan(ex0) else ex0 for ex0 in exog]
                x_mean, x_var = s['model'].predict(featureDict={'exog':exog_passed_in[0]})
            else:
                x_mean, x_var = s['model'].predict()
            x = [x_mean[0, 0]]
            x_std = [ math.sqrt( x_var[0,0] ) ]

            return x, x_std, s


    def dlm_set_exog_hyperparams(s, r:R_TYPE):
        # Univariate model with autoregressive components
        # This uses the discounting method of H/W so doesn't need to be fit as often
        period_choices = [3,5,7,10,12,16,24,32]
        s['auto_degree'], s['trend_degree'], period_choice = to_int_log_space(r,bounds=[
            (0.5,3.2),    # auto-reg degree. 1 means linear trend, 2 quadratic
            (0.1,3.4),    # seasonality degree
            (0,len(period_choices)) # seasonality period
        ])
        s['discount'] = 1 - ( r % 0.05 ) # Shouldn't matter too much as these get tuned eventually
        s['period'] = period_choices[period_choice]
        s['n_burn'] = 100
        s['n_fit']  = 500  # How often to tune discounts subsequently
        return s



if __name__ == '__main__':
    import matplotlib.pyplot as plt
    err = prior_plot_exogenous(f=dlm_exogenous_r3, k=1, n=1000, r=np.random.rand())
    plt.figure()
    print('done')
    pass



from timemachines.skatertools.visualization.priorplot import prior_plot_exogenous
try:
  from statsmodels.tsa.arima.model import ARIMA
except ImportError:
  from statsmodels.tsa.arima_model import ARIMA
from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any
from timemachines.skaters.tsa.tsaparams import TSA_META
from timemachines.skatertools.ensembling.ensemblefactory import precision_weighted_ensemble_factory
from timemachines.skatertools.components.parade import parade
from timemachines.skatertools.utilities.nonemath import nonecast
import warnings
from statsmodels.tools.sm_exceptions import ConvergenceWarning
warnings.simplefilter('ignore', ConvergenceWarning)


TSA_P_DEFAULT = 3
TSA_D_DEFAULT = 0
TSA_Q_DEFAULT = 3


def tsa_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None,
                             p:int=TSA_P_DEFAULT, d:int=TSA_D_DEFAULT, q:int=TSA_D_DEFAULT) -> ([float], Any, Any):
    """ Extremely simple univariate, fixed p,d,q ARIMA model that is re-fit each time """

    # TODO: FIX THIS TO USE EMPIRICAL STD, OTHERWISE ENSEMBLES ARE DREADFUL

    y = wrap(y)
    a = wrap(a)

    if not s.get('y'):
        s = {'y': list(),
             'a': list(),
             'k': k,
             'p':{}}
    else:
        # Assert immutability of k, dimensions
        if s['y']:
            assert len(y) == len(s['y'][0])
            assert k == s['k']
        if s['a']:
            assert len(a) == len(s['a'][0])

    if y is None:
        return None, s, None
    else:
        s['y'].append(y)
        if a is not None:
            s['a'].append(a)
        if len(s['y']) > max(2 * k + 5, TSA_META['n_warm']):
            y0s = [ y_[0] for y_ in s['y']]
            model = ARIMA(y0s, order=(p,d,q))
            try:
                x = list( model.fit().forecast(steps=k) )
            except:
                x = [wrap(y)[0]]*k
        else:
            x = [y[0]] * k

        y0 = wrap(y)[0]
        _we_ignore_bias, x_std, s['p'] = parade(p=s['p'], x=x, y=y0)
        x_std_fallback = nonecast(x_std, fill_value=1.0)
        return x, x_std_fallback, s




def tsa_p1_d0_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=1,d=0,q=0)


def tsa_p2_d0_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=2,d=0,q=0)


def tsa_p3_d0_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=3,d=0,q=0)

TSA_D0_Q0_SKATERS = [ tsa_p1_d0_q0, tsa_p2_d0_q0,
                      tsa_p3_d0_q0 ]

def tsa_p1_d0_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=1,d=0,q=1)


def tsa_p2_d0_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=2,d=0,q=1)


def tsa_p3_d0_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=3,d=0,q=1)

TSA_D0_Q1_SKATERS = [ tsa_p1_d0_q1, tsa_p2_d0_q1,
                      tsa_p3_d0_q1 ]

TSA_D0_SKATERS = TSA_D0_Q0_SKATERS + TSA_D0_Q1_SKATERS

#################################### d=1 ####################################



def tsa_p1_d1_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=1,d=1,q=0)


def tsa_p2_d1_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=2,d=1,q=0)


def tsa_p3_d1_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=3,d=1,q=0)


TSA_D1_Q0_SKATERS = [ tsa_p1_d1_q0, tsa_p2_d1_q0,
                      tsa_p3_d1_q0 ]

def tsa_p1_d1_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=1,d=1,q=1)


def tsa_p2_d1_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=2,d=1,q=1)


def tsa_p3_d1_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=3,d=1,q=1)


TSA_D1_Q1_SKATERS = [ tsa_p1_d1_q0, tsa_p2_d1_q0,
                      tsa_p3_d1_q0 ]

TSA_D1_SKATERS = TSA_D1_Q0_SKATERS + TSA_D1_Q1_SKATERS

TSA_CONSTANT_SKATERS = TSA_D0_SKATERS + TSA_D1_SKATERS





def tsa_balanced_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
         "less than precision" weighted moving averages
    """
    fs = TSA_CONSTANT_SKATERS
    return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.25)


def tsa_precise_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
         Precision weight moving averages
    """
    fs = TSA_CONSTANT_SKATERS
    return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.5)


def tsa_aggressive_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
         Precision weight moving averages
    """
    fs = TSA_CONSTANT_SKATERS
    return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.75)


def tsa_ma_precise_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
         Precision weight moving averages
    """
    from timemachines.skaters.simple.movingaverage import EMA_SKATERS
    fs = TSA_CONSTANT_SKATERS
    return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.5)


TSA_ENSEMBLE_SKATERS = [ tsa_aggressive_ensemble, tsa_balanced_ensemble, tsa_precise_ensemble ]

TSA_SKATERS = TSA_ENSEMBLE_SKATERS + TSA_CONSTANT_SKATERS


if __name__=='__main__':
    prior_plot_exogenous(f=tsa_p1_d0_q1,k=5,n=TSA_META['n_warm']+25,n_plot=50)



from timemachines.skaters.pmd.pmdinclusion import using_pmd
if using_pmd:
    from timemachines.skaters.pmd.pmdskaters import pmd_exogenous
    from timemachines.skatertools.data import hospital_with_exog
    from pmdarima.arima import ARIMA


    # So the task here is to serialize and re-inflate the arima model, and thus all the skater state, to JSON
    # It may be a lost cause, but by all means prove me wrong.


    def example_pmd():
        s = {}
        y, a = hospital_with_exog(k=3)
        x = [ pmd_exogenous(y=yj,s=s,k=3,a=aj) for yj,aj in zip(y[:500],a) ]
        return s


    def arima_res_to_dict(arima_res):
        state = arima_res.__dict__
        return state


    def pmd_to_dict(pmd):
        pmd['model'] = pmd['model'].__getstate__()
        pmd['model']['arima_res_'] = arima_res_to_dict(pmd['model']['arima_res_'])
        return pmd


    def pmd_from_dict(pmd):
        pmd['model']['arima_res_'] = ''


if __name__=='__main__':
    assert using_pmd,'pip install pmdarima'
    pmd = example_pmd()
    model = pmd['model']
    model1 = ARIMA(**model.get_params())
    prms = model.__dict__['arima_res_'].__dict__['_results'].params




